#!/usr/bin/perl

###############################################
# This script is used to Warp subject image
# to model image and using HAMMER. The output 
# is the deformation
# field from model to subject space, called
# "Field.1.img.Mdl2Subj", and a warped subject
# image(Warped.SubjSeg.img, already in 
# model space)
#
# 10/22/2008
# 
#
###############################################

use Getopt::Std;
use IO::File;
use Cwd;


#get arguments list
$argc=@ARGV;

if ($argc < 1){
   &showUsage ;
}

my $opt_string = 'hTS:O:M:i:R:r:m:c:X:x:E:D:P:V:v:d:';

getopts("$opt_string") || &showUsage ;

if($opt_h){
  &showUsage;
  }

#define argument list and get the arguments

#make a temp directory as a working directory:
if($opt_d){
    $tempDirectory="$opt_d";
}else{
    print "\n\tPlease provide a name which will be used to creat working directory for current data\n\n";
    exit(1);
}

  

# get Input files:
if($opt_S){
   $FileSegmented_orig=$opt_S;
}else{
   print "\n\n Must input a subject segmented image\n\n";
   &showUsage;
}

# get Input files:
if($opt_O){
   $FileOriginal_orig=$opt_O;
}else{
   print "\n\n Must input a subject original image\n\n";
   &showUsage;
}

if($opt_M){
   $MdlFile_orig =$opt_M;
}else{
   print "\n\n Must input a model segmented image\n\n";
   &showUsage;
}


#iteration number
$iteration = 50;
if($opt_i){
  $iteration=$opt_i;
}

# step ratio
$step_ration = 0.05;
if($opt_R){
  $step_ration=$opt_R;
}

# neighborhood search size, default: 12 for Low resolution, 
# 10 for Middl resolution 8 for High resolutions
@search_size = (12,10,8);
if($opt_r){
  $sizes = $opt_r;
  @search_size = split(",", $sizes);
}

# option: -m
$m_degree=0.6;
if($opt_m){
   $m_degree=$opt_m;
}


$sub_path = "./" ;

# option: -X
@subImage_size=(256,256,124);
if($opt_X){
  $sizes = $opt_X;
  @subImage_size=split(",", $sizes);
}

# option: -x
@modImage_size=(256,256,124);
if($opt_x){
  $sizes = $opt_x;
  @modImage_size=split(",", $sizes);
}

if(($subImage_size[0] != $subImage_size[1]) || ($modImage_size[0]!= $modImage_size[1]) || ($modImage_size[0] != $subImage_size[0]) ) {
    print "\nthe input images (subject image and model image) must have same size on X,Y plane !\n\n";
	exit(1);
	}

	
# option: -V
@subject_voxel=(0.9375,0.9375,1.5);
if($opt_V){
   $voxels=$opt_V;
   @subject_voxel = split(",", $voxels);
}

# option: -v
@model_voxel=(1,1,1);
if($opt_v){
   $voxels=$opt_v;
   @model_voxel = split(",", $voxels);
}


# option: -E
$E_SmthIter=1;
if($opt_E){
   $E_SmthIter=$opt_E;
}



print "\n\nLabel Brain Image by Fast HAMMER Version 1.0 @ IEDA LAB, UNC@Chapel Hill\n\n";
print "Input information: \n";
print "\t input subject segmented image is $FileSegmented_orig\n";
print "\t input subject original image is $FileOriginal_orig\n";
print "\t input Model image is $MdlFile_orig\n";
print "\t subect image size: $subImage_size[0], $subImage_size[1], $subImage_size[2]\n";
print "\t subject image resolution: $subject_voxel[0], $subject_voxel[1], $subject_voxel[2]\n";
print "\t model image resolution: $model_voxel[0], $model_voxel[1], $model_voxel[2]\n";
print "\t model image size: $modImage_size[0], $modImage_size[1], $modImage_size[2]\n";
print "\t iteration number: $iteration\n";
print "\t step ratio: $step_ration\n";
print "\t neighborhood search size: ($search_size[0], $search_size[1], $search_size[2])\n";
print "\t similarity degree: $m_degree\n";


# option: -T
if($opt_T){
   print "\tDo not output temporary warping results\n";
   $temp_output="-T";
}else{
   print "\tOutput temporary warping results\n";
   $temp_output="";
}

print "\n\n\n";

$original_path=getcwd;
print "Current directory: $original_path\n";

open (IN, "$FileSegmented_orig") || die "\nError open: $FileSegmented_orig\n\n";
close(IN);
open (IN, "$FileOriginal_orig") || die "\nError open: $FileOriginal_orig\n\n";
close(IN);
open (IN, "$MdlFile_orig") || die "\nError open: $MdlFile_orig\n\n";
close(IN);

print "The temporary results will be written in $tempDirectory\n\n";

#if ( -e $tempDirectory ) {
#    print "$tempDirectory exists! please provide a different name of 
#temporary working directory: option -d \n\n";
#    exit(1);
#} {
    print "mkdir $tempDirectory\n";
    mkdir($tempDirectory, 0777) ;
#||  die "Can't create $tempDirectory  directory\n";
#}
 

$FileSegmented=$tempDirectory."FileSegmented.img";
$File_Rigid=$FileSegmented.".rig";
$File_Rigid_name=$File_Rigid;

$FileOriginal=$tempDirectory."FileOrignal";
$FileOriginal_Rigid=$FileOriginal.".rig";
$FileOriginal_Rigid_name=$FileOriginal_Rigid;

$MdlFile="MdlFile.img";


$tempFileSegmented=$tempDirectory."\\".$FileSegmented;
$tempMdlFile=$tempDirectory."\\".$MdlFile;
$tempFileOriginal=$tempDirectory."\\".$FileOriginal;

print ("copy $FileSegmented_orig $tempFileSegmented\n");
system("copy $FileSegmented_orig $tempFileSegmented") && die "Can't copy file $FileSegmented_orig\n";
print ("copy $MdlFile_orig $tempMdlFile\n");
system("copy $MdlFile_orig $tempMdlFile") && die "Can't copy file $MdlFile_orig\n" ;
print ("copy $FileOriginal_orig $tempFileOriginal\n");
system("copy $FileOriginal_orig $tempFileOriginal") && die "Can't copy file $FileOriginal_orig\n" ;

chdir($tempDirectory);

open (IN, "$FileSegmented") || die "\nError open: $FileSegmented at $tempDirectory\n\n";
close(IN);
open (IN, "$MdlFile") || die "\nError open: $MdlFile at $tempDirectory\n\n";
close(IN);


#check image size if subject z-size is larger than template image, padding tempalte image
if ( $subImage_size[2] > $modImage_size[2] ){
    $newSize=$subImage_size[2] - 1;
    $command="CropSlices";
    print "$command $MdlFile $MdlFile -c0,$newSize\n";
    system("$command $MdlFile $MdlFile -c0,$newSize") && die "Error: Padding model Image, please check the model image: $MdlFile_orig, size, data format\n";
}

# affine transformation: for upsample MdlFile
$command="LinearlyRegisterImgA2ImgB-byFSL.exe";
print "\n\nLinear align two images\n";
print "\n$command  $FileSegmented $MdlFile $File_Rigid $FileOriginal 12 $subject_voxel[0] $subject_voxel[1] $subject_voxel[2] $model_voxel[0] $model_voxel[1] $model_voxel[2] $subImage_size[0]\n";
#system ("$command");
system("$command  $FileSegmented $MdlFile $File_Rigid $FileOriginal 12 $subject_voxel[0] $subject_voxel[1] $subject_voxel[2] $model_voxel[0] $model_voxel[1] $model_voxel[2] $subImage_size[0]") && die "Error: Linear align two images, check the FSL package  was installed properly and the input images\n";
print "Done rigid transformation\n\n";

$tempFile="temp_nouse.txt";
$HGH_FnlID=$iteration -1;

open (IN, "$File_Rigid") || die "\nError: program failed at:LinearlyRegisterImgA2ImgB-byFSL.exe, missing output: $File_Rigid.\n\n";
close(IN);
open (IN, "$FileOriginal_Rigid") || die "\nError: program failed at:LinearlyRegisterImgA2ImgB-byFSL.exe, missing output $FileOriginal_Rigid.\n\n";
close(IN);




print "\n"; 
print "*******************************************************************\n";
print "high resoultion:\n";
print "*******************************************************************\n";
print "Segmenting '$File_Rigid' in high-resolution\n";

$command="sysHAMMER.exe";
print "$command $MdlFile $File_Rigid_name dummy -R $step_ration,$step_ration,$step_ration -D $sub_path -i $iteration,$iteration,$iteration -r $search_size[0],$search_size[1],$search_size[2] -m $m_degree -c 0.5 -O $FileOriginal_Rigid_name -F -X $subImage_size[0],$subImage_size[1],$subImage_size[2] -p 2,1\n"; 
system("$command $MdlFile $File_Rigid_name dummy -R $step_ration,$step_ration,$step_ration -D $sub_path -i $iteration,$iteration,$iteration -r $search_size[0],$search_size[1],$search_size[2] -m $m_degree -c 0.5 -O $FileOriginal_Rigid_name -F -X $subImage_size[0],$subImage_size[1],$subImage_size[2] -p 2,1");


print "\nRegistration Done!\n\n";

#******************************************************
# generate deformation field from subject to model
#******************************************************
$fieldFile=$tempDirectory.".Field.1.img.Mdl2Subj";
#$command="splitfile";
#print "\n\n$command DeformationField.float.img NoUse 8 Field.1.img\n";
#system("$command DeformationField.float.img NoUse 8 Field.1.img");
$command="Transform3DVectorField";
print "\n$command dummy $fieldFile -v $subImage_size[0],$subImage_size[1],$subImage_size[2]  \n";
system("$command dummy $fieldFile -v $subImage_size[0],$subImage_size[1],$subImage_size[2]  ");

#***************************************************
# crop subject image to model space if model z-dimension is larger than subject
#***************************************************

$croppedFileSegmented=$FileSegmented."cropped";
$croppedFileOriginal=$FileOriginal."cropped";

if ( $subImage_size[2] < $modImage_size[2] ){
$command="CropSlices";
$SliceNo=$modImage_size[2]-1;

print "\n$command $FileSegmented $croppedFileSegmented  -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo\n";
system("$command $FileSegmented $croppedFileSegmented -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo");
print "\n$command $FileOriginal $croppedFileOriginal -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo\n";
system("$command $FileOriginal $croppedFileOriginal  -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo");
} else {

print "copy $FileSegmented $croppedFileSegmented\n";
system("copy $FileSegmented $croppedFileSegmented") && die "can't copy file $FileSegmented " ;

print "copy $FileOriginal $croppedFileOriginal\n";
system("copy $FileOriginal $croppedFileOriginal") && die "can't copy file $FileOriginal ";

}

#***************************************************
# generate warped subject segmented file
#***************************************************
$warpedSeg=$tempDirectory.".Warped.SubjSeg.img";
$command="PerformDeformationOnImgUsingVectorField";
print "\n$command  $fieldFile $croppedFileSegmented $warpedSeg -v $subImage_size[0],$subImage_size[1],$subImage_size[2]  -I\n\n";
system("$command  $fieldFile $croppedFileSegmented $warpedSeg -v $subImage_size[0],$subImage_size[1],$subImage_size[2]  -I");

#***************************************************
# generate warped model label file
#***************************************************
$warpedOriginal=$tempDirectory.".Warped.SubjOrig.img";
$command="PerformDeformationOnImgUsingVectorField";
print "\n$command $fieldFile $croppedFileOriginal $warpedOriginal -v $subImage_size[0],$subImage_size[1],$subImage_size[2]\n\n";
system("$command  $fieldFile $croppedFileOriginal $warpedOriginal -v $subImage_size[0],$subImage_size[1],$subImage_size[2]");

#***************************************************
# crop subject image to model space if the model image's z-dimension is less than subject's
#***************************************************
if ( $subImage_size[2] >  $modImage_size[2] ){
print "model image z-dimension is less than subject image's. Cropped the warped image. \n";
$command="CropSlices";
$SliceNo=$modImage_size[2]-1;
print "\n$command $warpedSeg $warpedSeg   -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo\n";
system("$command $warpedSeg $warpedSeg   -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo");
print "\n$command $warpedOriginal $warpedOriginal  -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo\n";
system("$command  $warpedOriginal $warpedOriginal  -v $subImage_size[0],$subImage_size[1],$subImage_size[2] -c 0,$SliceNo");
}


#***************************************************
# move results file to original directory
#***************************************************
system("move $fieldFile $original_path/.");
system("move $warpedSeg $original_path/.");
system("move $warpedOriginal $original_path/.");

#***************************************************
# remove unused files
#***************************************************
system("del $File_Rigid_name");
system("del $MdlFile");
system("del $LabelFile");
system("del $FileSegmented");
system("del $FileOriginal");
system("del $croppedFileSegmented");
system("del $croppedFileOriginal");

chdir($original_path);
print "Results are $fieldFile  $warpedSeg  $warpedOriginal\n";
print "Done !\n";

if($opt_T){
    system ("rmdir $tempDirectory /S /Q");
}else {
    print "\nTemporary file if you choice to be kept is in the directory: $tempDirectory\n";
}

# *************************************************
# function:  showUsage
# arguments: None
# *************************************************
sub showUsage{
    print "\n";
	print "This program is used to warped subject image to model space\n";
	print "output is the deformation field from model to subject(Field.1.img.Mdl1Subj)\n";
	print "and a warped subject image (Warped.SubjSeg.img), already in model space\n\n";
	
	print "Usage: $0 [options]\n";
	
	print "\t-h                          : display this (help) message\n";
	print "\t-S <string>                 : subject segmented image file name(including absolute path, if not in current working directory)\n";
	print "\t-O <string>                 : subject original image file name(including absolute path, if not in current working directory)\n";
	print "\t-d <string>                 : name of temporary directory(Please PROVIDE), program will be run from there\n\n";

	print "\t-M <string>                 : model segmented image file name(including absolute path, if not in current working directory)\n";
	print "\t-X <int, int, int>          : subject image(source) size, default: 256, 256,124\n";
	print "\t-x <int, int, int>          : model image(target) size, default: 256, 256,124\n";
	print "\t-V <float, float, float>    : subject image (source) resolution in X, Y, Z, default: 0.9375, 0.9375, 1.5\n";
	print "\t-v <float, float, float>    : model image (target) resolution in X, Y, Z, default: 1.0, 1.0, 1.0\n\n";

	print "\t-i <int>                    : iteration number, default: 50\n";
	print "\t-R <float>                  : step ratio, default: 0.05\n";
	print "\t-r <int, int, int>          : neighborhood search size, different at each resolution.\n"; 
	print "\t                              input as < value_at_low, value_at_middle, value_at_high>\n";
	print "\t                              default: <12, 10, 8> mm/voxels\n";
	print "\t-m <float>                  : the similarity degree, default: 0.6\n";
	print "\t-T                          : Don't output temporary warping results\n";
	print "\t-E <int>                    : edge-preserving smoothing; iteration number, default: 1\n";
	
	print "\nNotes:\n";
	print "\t          The program can only take byte type input imageas\n";
	print "\t          The X, Y dimension size of input image must be equal.\n";

    print "Example:\n"; 
	print "\tusing default value: \n";
	print "\t\t$0 -S subject -O subject_original_image -M model -d t1 \n\n";
	print "\twith options:\n";
	print "\t\t$0 -S subject_absolute_path/t1_segmented.img -O subject_absolute_path/t1_stripped.img -d t1 -M model_absolute_path/jakob_rad_convention_segmented.img -X256,256,110 -x256,256,198 -V0.9375,0.9375,1.5 -v1,1,1 -T\n\n";

    print "References:\n";
	print "\t1. Shen, D. and C. Davatzikos, \"Very high-resolution morphometry using mass-preserving\n";
	print "\t   deformations and HAMMER elastic registration\". Neuroimage, 2003. 18(1): p. 28-41.\n";
	print "\t2. Shen, D. and C. Davatzikos, \"HAMMER: Hierarchical attribute matching mechanism\n";
	print "\t   for elastic registration\". IEEE Trans on Med. Imaging, 2002. 21(11): p. 1421-1439.\n\n";
    print "\n\n";

	exit(1);
} 

